use std::intrinsics::{volatile_load, volatile_store};
use super::{GraphicLogger, Scene, gpu_setres_raw};

pub struct ClockScene {
}

fn clock_lambda<F>(f: F) -> u32 where F: FnOnce() {
	let timer0value = 0x1F801100 as *const u32;
	let timer0mode = 0x1F801104 as *mut u32;
	//let timer0target = 0x1F801108 as *mut u32;

	unsafe { volatile_store(timer0mode, 0); }
	let t1 = unsafe { volatile_load(timer0value) };
	f();
	let t2 = unsafe { volatile_load(timer0value) };

	return t2 - t1;
}

impl ClockScene {
#[cfg_attr(not(release), inline(never))]
	pub fn new(logger: &mut GraphicLogger) -> ClockScene {
		gpu_setres_raw(1, 0, 0, 0);

		logger.writeln(format!("null by timer0: {}", clock_lambda(||{})).as_bytes());

		logger.writeln(format!("multu by timer0: {}", clock_lambda(||unsafe{
			let v1 = 0xFFFFFFFFu32;
			let v2 = 0xFFFFFFFFu32;
			let v3: u32;
			asm!("multu $1, $2; mflo $0" : "=r"(v3) : "r"(v1), "r"(v2) : : "volatile");
		})).as_bytes());

		logger.writeln(format!("divu by timer0: {}", clock_lambda(||unsafe{
			let v1 = 0xFFFFFFFFu32;
			let v2 = 0xFFF00FFFu32;
			let v3: u32;
			asm!("divu $1, $2; mflo $0" : "=r"(v3) : "r"(v1), "r"(v2) : : "volatile");
		})).as_bytes());

		ClockScene {}
	}
}

impl Scene for ClockScene {
#[cfg_attr(not(release), inline(never))]
	fn draw(&self) {
		// draw nothing... done in new(!)
	}
#[cfg_attr(not(release), inline(never))]
	fn input(&mut self, _bits: u16) {
		// do nothing...
	}
}
