#![feature(core_intrinsics, libc, asm, const_fn)]

// note: Xebra implements load-delay-slot, while NO$PSX does not.
//       LLVM must be patched to handle load-delay-slot.

// TODO remove #[cfg_attr(not(release), inline(never))]

use std::intrinsics::{volatile_load, volatile_store};
use std::cmp;

//#[inline(always)]
//pub fn roundup_div(n: usize, d: usize) -> usize {
/*
pub fn roundup_div<T>(n: T, d: T) -> <<T as std::ops::Add>::Output as std::ops::Div>::Output
where T: std::ops::Sub + std::ops::Add + std::ops::Div,
      <T as std::ops::Add>::Output: std::ops::Div
      {
	return (n + (d - 1)) / d;
}
*/
// TODO what type does roundup_div should?
macro_rules! roundup_div {
	($n:expr, $d:expr) => (($n + ($d - 1)) / $d);
}
macro_rules! roundup {
	($n:expr, $d:expr) => (roundup_div!($n, $d) * $d);
}

pub mod sys; // FIXME squelching some "unused" warning (including for libc ex. memcpy that is correct)
#[macro_use]
#[allow(dead_code)] // FIXME squelching some "unused" warning (atan2)
mod fastmath;
#[macro_use]
#[allow(dead_code)] // FIXME squelching some "unused" warning
mod mat;
#[allow(dead_code)] // FIXME squelching some "unused" warning (PlanB)
mod graphic;
mod progress;
//mod testscene;
mod flappyscene;
//mod clockscene;
//mod pslogoscene;
//mod td1scene;
mod shootingscene;
mod titlescene;

pub mod prng {
	pub struct XorShift32 {
		state: u32,
	}
	impl XorShift32 {
#[cfg_attr(not(release), inline(never))]
		pub const fn new() -> XorShift32 {
			XorShift32 { state: 2463534242 }
		}
#[cfg_attr(not(release), inline(never))]
		pub fn seed(&mut self, seed: u32) {
			self.state = seed;
		}
#[cfg_attr(not(release), inline(never))]
		pub fn rand(&mut self) -> u32 {
			self.state = self.state ^ u32::wrapping_shl(self.state, 13);
			self.state = self.state ^ self.state >> 17;
			self.state = self.state ^ u32::wrapping_shl(self.state, 5);
			self.state
		}
	}
	// note: single-threaded! omitting wrapping with Mutex
	static mut g_xorshift: XorShift32 = XorShift32::new();
	pub fn rand() -> u32 { unsafe { g_xorshift.rand() } }
#[cfg_attr(not(release), inline(never))]
	pub fn randn(max: u32) -> u32 { ((rand() as u64 * max as u64) >> 32) as u32 }
	pub fn seed(seed: u32) { unsafe { g_xorshift.seed(seed) } }
}

#[cfg_attr(not(release), inline(never))]
pub fn gp0(w: u32) {
	unsafe {
		// wait until ready to accept gp0cmd
		//while volatile_load(0x1F801814 as *const u32) & 0x0400_0000 == 0 {}
		// then send cmd
		volatile_store(0x1F801810 as *mut u32, w);
	}
}

#[cfg_attr(not(release), inline(never))]
pub fn gp1(w: u32) {
	unsafe { volatile_store(0x1F801814 as *mut u32, w); }
}

#[cfg_attr(not(release), inline(never))]
pub fn gpustat() -> u32 {
	unsafe { return volatile_load(0x1F801814 as *mut u32); }
}

#[cfg_attr(not(release), inline(never))]
pub fn set_draw_area(topleft: (u32, u32), bottomright: (u32, u32)) {
	gp0(0xE3000000 | (topleft.0     << 0) | (topleft.1     << 10)); // draw(clip)x1y1
	gp0(0xE4000000 | (bottomright.0 << 0) | (bottomright.1 << 10)); // draw(clip)x2y2
	gp0(0xE5000000 | (topleft.0     << 0) | (topleft.1     << 11)); // drawoffset (note: 11bit.)
}

#[cfg_attr(not(release), inline(never))]
pub fn spustat_waitready() {
	let spustat = 0x1F801DAE as *mut u16;
	while unsafe { volatile_load(spustat) } & 0x0400 != 0 { /* do nothing */ }
}

#[cfg_attr(not(release), inline(never))]
pub fn spucnt_xfermode(mode_raw: u16) {
	let p_spucnt = 0x1F801DAA as *mut u16;
	let p_spustat = 0x1F801DAE as *mut u16;

	assert!(mode_raw & 0x0003 == mode_raw);
	let mode = (mode_raw & 0x0003) << 4;
	let mode_mask = 0x0030;

	// write same value is used by spu_fifoload
	/*
	let chkv = unsafe { volatile_load(p_spustat) };

	if chkv & mode_mask == mode {
		return;
	}
	*/

	let prev = unsafe { volatile_load(p_spucnt) };
	let newv = (prev & !mode_mask) | mode;
	unsafe { volatile_store(p_spucnt, newv) };

	loop {
		let curv = unsafe { volatile_load(p_spustat) };
		if curv & mode_mask == mode {
			return;
		}
	}
}

fn dma_base(ch: u8) -> u32 {
	if 7 <= ch {
		panic!("Invalid DMA channel specified");
	}
	return 0x1F801080 + 0x10 * ch as u32;
}

#[cfg_attr(not(release), inline(never))]
pub fn dmawait(ch: u8) {
	let dmacctl = (dma_base(ch) + 8) as *mut u32; // channel control
	loop {
		let ctl = unsafe { volatile_load(dmacctl) };
		if ctl & 0x01000000 == 0 {
			return;
		}
	}
}

// FIXME should unsafe fn, because this cannot ensure safety. ptr+length -> &[T]?
// note: length * blocksize is in words(4bytes), not in bytes.
#[cfg_attr(not(release), inline(never))]
pub fn dma(ch: u8, ptr: *const u8, length: u16, blocksize: u16, ctl: u32) {
	assert!(ptr as u32 & 3 == 0, "dma ptr must be aligned by word");

	let base = dma_base(ch);
	let dmaaddr = (base + 0) as *mut u32;
	let dmabcr = (base + 4) as *mut u32; // block control
	let dmacctl = (base + 8) as *mut u32; // channel control
	// ensure target channel enabled, with priority 1 (by BIOS MDECin,GPU,CDROM are enabled with priority all 1)
	// note: bootscreen's libspu sets SPU-DMA priority to 3? (but it seems it does not use DMA...)
	unsafe {
		let dmadpcr = 0x1F8010F0 as *mut u32;
		volatile_store(dmadpcr, volatile_load(dmadpcr) & !(0xF << (ch * 4)) | (0x9 << (ch * 4)));
	}
	//dmawait(ch); // wait until last DMA completes on this channel
	unsafe {
		volatile_store(dmaaddr, ptr as u32 & 0x00FFffff);
		volatile_store(dmabcr, ((length as u32) << 16) | blocksize as u32);
		volatile_store(dmacctl, ctl);
	}
}

#[cfg_attr(not(release), inline(never))]
pub fn spucnt_mask_set(mask: u16, set: u16) {
	let spuctl = 0x1F801DAA as *mut u16;
	unsafe { volatile_store(spuctl, (volatile_load(spuctl) & mask) | set) };
}

#[cfg_attr(not(release), inline(never))]
pub fn spucnt_enable(enable: bool) {
	spucnt_mask_set(0x7FFF, if enable { 0x8000 } else { 0 });
}

#[cfg_attr(not(release), inline(never))]
pub fn spucnt_unmute(unmute: bool) {
	spucnt_mask_set(0xBFFF, if unmute { 0x4000 } else { 0 });
}

// note: enable reverb master AFTER setting mBase or your ADPCM samples go away!
#[cfg_attr(not(release), inline(never))]
pub fn spucnt_reverb(enable: bool) {
	spucnt_mask_set(0xFF7F, if enable { 0x0080 } else { 0 });
}

// note: address in 8bytes block.
#[cfg_attr(not(release), inline(never))]
pub fn spucnt_xfertype_addr(xfertype: u16, addr: u16) {
	unsafe { volatile_store(0x1F801DAC as *mut u16, xfertype) };
	spucnt_xfermode(0); // stop
	unsafe { volatile_store(0x1F801DA6 as *mut u16, addr) };
}

#[cfg_attr(not(release), inline(never))]
pub fn spu_dmaload<T: AsRef<[u8]>>(dstaddr: u32, vb_r: T) {
	let vb = vb_r.as_ref();
	if vb.len() < 8 {
		panic!("Can't load to SPU memory that is smaller than 8 bytes");
	}
	// enable spu, or Xebra stops on dmawait (NO$PSX continues that looks as an emulation glitch)
	spucnt_enable(true);
	spucnt_xfertype_addr(0x0004, (dstaddr >> 3) as u16); // xfertype 2: normal
	spucnt_xfermode(2); // dmawrite

	spustat_waitready();

	let blocklog2: u8 = 2; // 2^2 = 4 words at once.
	// DMA4 block=0x10 ctl=0x01000201: fromMem, no-chopping, mode1, start
	dma(4, vb.as_ptr(), ((vb.len() / std::mem::size_of::<u32>()) >> blocklog2) as u16, 1 << blocklog2, 0x01000201);
	dmawait(4);
	spustat_waitready(); // FIXME required?
}

#[cfg_attr(not(release), inline(never))]
pub fn spu_fifoload(dstaddr: u32, vb: &[u16]) {
	let spufifo = 0x1F801DA8 as *mut u16;

	// enable spu for convenience.
	// TODO if xfering by fifo, Xebra also stops?
	spucnt_enable(true);
	spucnt_xfertype_addr(0x0004, (dstaddr >> 3) as u16); // xfertype 2: normal

	// libspu in bootscreen waits after each 64 bytes (32 halfwords) write.
	for chunk in vb.chunks(32) {
		for hw in chunk {
			unsafe { volatile_store(spufifo, *hw); }
		}
		spucnt_xfermode(1); // fifowrite
		spustat_waitready();
	}
}

/*
#[derive(Clone)]
pub struct SpuParam {
	vol: (u16, u16),
	srate: u32,
	addr: u32,
	env: u32, // adsr, renamed because confuse with addr
}

impl Default for SpuParam {
	fn default() -> SpuParam {
		return SpuParam {
			vol: (0x4000, 0x4000),
			srate: 22050, // really?
			addr: 0x1000, // really?
			env: 0x000F5F00
		}
	}
}

#[cfg_attr(not(release), inline(never))]
pub fn spu_play_full(ch: u8, param: SpuParam) {
	let pvoll = (0x1F801C00 + 0x10 * ch as u32) as *mut u16;
	let pvolr = (0x1F801C02 + 0x10 * ch as u32) as *mut u16;
	let prate = (0x1F801C04 + 0x10 * ch as u32) as *mut u16;
	let paddr = (0x1F801C06 + 0x10 * ch as u32) as *mut u16;
	let pads  = (0x1F801C08 + 0x10 * ch as u32) as *mut u16;
	let pr    = (0x1F801C0A + 0x10 * ch as u32) as *mut u16;
	// 0x1F801C0C,E are adsr counter etc, almost read only.
	let p_kon  = 0x1F801D88 as *mut u16; // TODO can this accessed by 32bit? (seems good on NO$PSX)
	let r16 = (param.srate * 0x1000 / 44100) as u16; // 0x1000 means 44100Hz
	unsafe {
		volatile_store(pvoll, param.vol.0);
		volatile_store(pvolr, param.vol.1);
		volatile_store(prate, r16);
		volatile_store(paddr, (param.addr >> 3) as u16);
		volatile_store(pads, (param.env >> 16) as u16);
		volatile_store(pr, (param.env & 0xFFFF) as u16);
		if ch < 16 {
			volatile_store(p_kon, 1 << ch);
		} else {
			volatile_store(p_kon.offset(1), 1 << (ch - 16));
		}
	}
}
*/

// TODO Is it good to merge vol as a u32, instead of u16 tuple?
//      Consider for both source-coding and binary.
#[cfg_attr(not(release), inline(never))]
pub fn spu_play_ex(ch: u8, addr: u32, srate: u32, vol: (u16, u16), env: u32, reverb: bool) {
	let pvoll = (0x1F801C00 + 0x10 * ch as u32) as *mut u16;
	let pvolr = (0x1F801C02 + 0x10 * ch as u32) as *mut u16;
	let prate = (0x1F801C04 + 0x10 * ch as u32) as *mut u16;
	let paddr = (0x1F801C06 + 0x10 * ch as u32) as *mut u16;
	let pads  = (0x1F801C08 + 0x10 * ch as u32) as *mut u16;
	let pr    = (0x1F801C0A + 0x10 * ch as u32) as *mut u16;
	// 0x1F801C0C,E are adsr counter etc, almost read only.
	let p_kon  = 0x1F801D88 as *mut u16; // only this is write-only-reg. TODO can this accessed by 32bit? (seems good on NO$PSX)
	let p_non  = 0x1F801D94 as *mut u16; // TODO can this accessed by 32bit?
	let p_eon  = 0x1F801D98 as *mut u16; // TODO can this accessed by 32bit?
	let srate16 = (srate * 0x1000 / 44100) as u16; // 0x1000 means 44100Hz
	unsafe {
		volatile_store(pvoll, vol.0);
		volatile_store(pvolr, vol.1);
		volatile_store(prate, srate16);
		volatile_store(paddr, (addr >> 3) as u16);
		volatile_store(pads, (env >> 16) as u16);
		volatile_store(pr, (env & 0xFFFF) as u16);
		if ch < 16 {
			volatile_store(p_non, volatile_load(p_non) & !(1 << ch)); // use ADPCM
			volatile_store(p_eon, volatile_load(p_eon) & !(1 << ch) | (if reverb { 1 } else { 0 } << ch));
			volatile_store(p_kon, 1 << ch);
		} else {
			volatile_store(p_non.offset(1), volatile_load(p_non.offset(1)) & !(1 << (ch - 16))); // use ADPCM
			volatile_store(p_eon.offset(1), volatile_load(p_eon.offset(1)) & !(1 << (ch - 16)) | (if reverb { 1 } else { 0 } << (ch - 16)));
			volatile_store(p_kon.offset(1), 1 << (ch - 16));
		}
	}
}

#[cfg_attr(not(release), inline(never))]
pub fn spu_play(ch: u8, addr: u32, srate: u32) {
	// 000F attack: linear,fastest,step0=+7 decay: fastest sustain: levelF
	// 5F00 sustain: linear,decrease,slowest,step0=-5 release: linear,fastest
	//spu_play_full(ch, SpuParam { addr: addr, srate: srate, ..Default::default() } );
	spu_play_ex(ch, addr, srate, (0x4000, 0x4000), 0x000F5F00, true);
}

#[cfg_attr(not(release), inline(never))]
pub fn spu_keyoff(ch: u8) {
	let p_koff = 0x1F801D8C as *mut u16; // TODO can this accessed by 32bit?
	unsafe {
		if ch < 16 {
			volatile_store(p_koff, 1 << ch);
		} else {
			volatile_store(p_koff.offset(1), 1 << (ch - 16));
		}
	}
}

#[cfg_attr(not(release), inline(never))]
pub fn spu_vol(main: (u16, u16), reverb: (u16, u16)) {
	let p_mainvoll = 0x1F801D80 as *mut u16;
	let p_mainvolr = 0x1F801D82 as *mut u16;
	let p_revvoll = 0x1F801D84 as *mut u16;
	let p_revvolr = 0x1F801D86 as *mut u16;
	unsafe {
		volatile_store(p_mainvoll, main.0);
		volatile_store(p_mainvolr, main.1);
		volatile_store(p_revvoll, reverb.0);
		volatile_store(p_revvolr, reverb.1);
	}
}

#[cfg_attr(not(release), inline(never))]
fn gp0rgb(cmd: u8, r: u8, g: u8, b: u8) {
	// 0xCcBbGgRr
	gp0(((cmd as u32) << 24) | ((r as u32) << 0) | ((g as u32) << 8) | ((b as u32) << 16));
}

#[cfg_attr(not(release), inline(never))]
pub fn fillvram(r: u8, g: u8, b: u8, x: u16, y: u16, w: u16, h: u16) {
	gp0rgb(0x02, r, g, b);
	gp0(((y as u32) << 16) | x as u32);
	gp0(((h as u32) << 16) | w as u32);
}

#[cfg_attr(not(release), inline(never))]
pub fn fillrect(r: u8, g: u8, b: u8, x: u16, y: u16, w: u16, h: u16) {
	gp0rgb(0x60, r, g, b);
	gp0(((y as u32) << 16) | x as u32);
	gp0(((h as u32) << 16) | w as u32);
}

// FIXME this fn blocks... async?
// TODO merge with gpu_load, so much duplicated.
#[cfg_attr(not(release), inline(never))]
pub fn gpu_byteload<T: AsRef<[u8]>>(x16: u16, y16: u16, w16: u16, h16: u16, data_r: T) {
	let data = data_r.as_ref();
	let x = x16 as u32;
	let y = y16 as u32;
	let w = w16 as u32;
	let h = h16 as u32;

	let gp0 = 0x1F801810 as *mut u32;
	let blkw = 1;

	assert!((w as usize * h as usize) <= data.len(), "data must be same or larger than width*height");

	assert!(w % 2 == 0, "width must be even to make word-alignment"); // FIXME only because lazyness. not tested.
	assert!(roundup_div!(w, 2) % blkw == 0, "width must be multiply of block-words"); // FIXME only because lazyness. complex.

	// VRAM is 1024x512.
	assert!(x < 1024, "position x must in VRAM area");
	assert!(y < 512, "position y must in VRAM area");
	// note: w and h are exclusive end, so uses "<=" not "<".
	assert!(w <= 1024, "width must not bigger than VRAM area");
	assert!(h <= 512, "height must not bigger than VRAM area");
	assert!(x + w <= 1024, "pos.x + width must in VRAM area");
	assert!(y + h <= 1024, "pos.y + height must in VRAM area");

	// faster lazier bit-cryptic method
	let block_max = 0xFFFF * 2 * blkw; // 2=halfword per word. 0=0x1_0000 is allowed by DMA but Rust does not allow overflow.
	let lines_at_once = (|| { for i in 0.. { if block_max < (w << i) { return 1 << (i - 1) } }; return 0; })(); // bitshift is faster than div (but smaller block size)
	let blocks = roundup_div!(h, lines_at_once);
	for bi in 0..blocks {
		let lines_to_be_xfer = cmp::min(lines_at_once, h - lines_at_once * bi);

		unsafe {
			volatile_store(gp0, 0xA0000000); // cpu2vram
			volatile_store(gp0, (x << 0) | ((y + lines_at_once * bi) << 16));
			volatile_store(gp0, (w << 0) | (lines_to_be_xfer << 16));
		}

		let ptr = unsafe { data.as_ptr().offset((w * lines_at_once * bi) as isize) } as *const u8;
		dma(2, ptr, (w * lines_to_be_xfer / 2 / blkw) as u16, blkw as u16, 0x01000201);
		dmawait(2);
	}
}

// FIXME is [u16] guarantees only half-word align that is bad?
//       in practice, malloc is aligned by word, that is good for DMA.
// FIXME this fn blocks... async?
#[cfg_attr(not(release), inline(never))]
pub fn gpu_load(x16: u16, y16: u16, w16: u16, h16: u16, data: &[u16]) {
	let x = x16 as u32;
	let y = y16 as u32;
	let w = w16 as u32;
	let h = h16 as u32;

	let gp0 = 0x1F801810 as *mut u32;
	let blkw = 1;

	assert!((w as usize * h as usize) <= data.len(), "data must be same or larger than width*height");

	assert!(w % 2 == 0, "width must be even to make word-alignment"); // FIXME only because lazyness. not tested.
	assert!(roundup_div!(w, 2) % blkw == 0, "width must be multiply of block-words"); // FIXME only because lazyness. complex.

	// VRAM is 1024x512.
	assert!(x < 1024, "position x must in VRAM area");
	assert!(y < 512, "position y must in VRAM area");
	// note: w and h are exclusive end, so uses "<=" not "<".
	assert!(w <= 1024, "width must not bigger than VRAM area");
	assert!(h <= 512, "height must not bigger than VRAM area");
	assert!(x + w <= 1024, "pos.x + width must in VRAM area");
	assert!(y + h <= 1024, "pos.y + height must in VRAM area");

	/*
	let divword = 0x80000;

	let words_per_line = roundup_div!(w, 2);
	assert!(words_per_line < divword, "it must at least one line per div"); // FIXME only because lazyness. too complex.
	let lines_per_div = divword / words_per_line;
	let loops = roundup_div!(h, lines_per_div);

	for di in 0..loops {
		unsafe {
			volatile_store(gp0, 0xA0000000); // cpu2vram
			volatile_store(gp0, (x << 0) | ((y + lines_per_div * di) << 16));
			volatile_store(gp0, (w << 0) | (lines_per_div << 16));

			let ptr = data.as_ptr().offset((w * lines_per_div * di) as isize) as *const u8;
			// following DMAs are not required to be in unsafe{}, but ptr is.
			dma(2, ptr, (w * lines_per_div / 2 / blkw) as u16, blkw as u16, 0x01000201);
			dmawait(2);
		}
	}
	*/

	// faster lazier bit-cryptic method
	let block_max = 0xFFFF * 2 * blkw; // 2=halfword per word. 0=0x1_0000 is allowed by DMA but Rust does not allow overflow.
	let lines_at_once = (|| { for i in 0.. { if block_max < (w << i) { return 1 << (i - 1) } }; return 0; })(); // bitshift is faster than div (but smaller block size)
	let blocks = roundup_div!(h, lines_at_once);
	for bi in 0..blocks {
		let lines_to_be_xfer = cmp::min(lines_at_once, h - lines_at_once * bi);

		unsafe {
			volatile_store(gp0, 0xA0000000); // cpu2vram
			volatile_store(gp0, (x << 0) | ((y + lines_at_once * bi) << 16));
			volatile_store(gp0, (w << 0) | (lines_to_be_xfer << 16));
		}

		let ptr = unsafe { data.as_ptr().offset((w * lines_at_once * bi) as isize) } as *const u8;
		dma(2, ptr, (w * lines_to_be_xfer / 2 / blkw) as u16, blkw as u16, 0x01000201);
		dmawait(2);
	}
}

#[cfg_attr(not(release), inline(never))]
pub fn read_digipad() -> u16 {
	// FIXME isn't there volatile variable in rust? or can't I write as v_load(buf)?
	if unsafe { volatile_load((&sys::padbuf[0][1]) as *const u8) } != 0x41 {
		return 0; // that is not digipad; report as none pressed.
	}
	// 0x41: digital pad
	// note: 0 if downed, 1 if released (reverse of natural!), so reverting bits here.
	return ! unsafe {
		((volatile_load((&sys::padbuf[0][3]) as *const u8) as u16) << 8) |
			(volatile_load((&sys::padbuf[0][2]) as *const u8) as u16) };
}

#[cfg_attr(not(release), inline(never))]
pub fn gp1disprange(x1: u32, x2: u32, y1: u32, y2: u32) {
	assert!(x1 & 0xFFF == x1, "x1 must be in 0 <= x1 <= 0xFFF");
	assert!(x2 & 0xFFF == x2, "x2 must be in 0 <= x2 <= 0xFFF");
	assert!(y1 & 0x3FF == y1, "y1 must be in 0 <= y1 <= 0x3FF");
	assert!(y2 & 0x3FF == y2, "y2 must be in 0 <= y2 <= 0x3FF");
	gp1(0x06000000 | (x1 << 0) | (x2 << 12)); // horz disprange; x1,x2=12bit
	gp1(0x07000000 | (y1 << 0) | (y2 << 10)); // vert disprange; y1,y2=10bit
}

#[cfg_attr(not(release), inline(never))]
pub fn calc_gpu_disprange(wpix: u32, h: u32) -> (u32, u32, u32, u32) {
	// note: vert disprange constant for NTSC
	return (0x260, 0x260 + wpix, 0x88 - (h / 2), 0x88 + (h / 2));
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_disprange(wpix: u32, h: u32) {
	let (x1, x2, y1, y2) = calc_gpu_disprange(wpix, h);
	gp1disprange(x1, x2, y1, y2);
}

#[cfg_attr(not(release), inline(never))]
pub fn calc_gpu_res_params(w: u32, h: u32, bpp: u32, rate: u32) -> (u32, u32, u32, u32, u32) {
	let horz = [256 * 10, 320 * 8, 512 * 5, 640 * 4]; // all are same value = 2560...
	// note: using h as vert and also interlace.
	let gp1cmd = 0x08000000 | (h << 5) | (bpp << 4) | (rate << 3) | (h << 2) | w;
	// note: h is a height-at-a-frame, which is same for 240p and 480i.
	let (x1, x2, y1, y2) = calc_gpu_disprange(horz[w as usize], 240);
	return (gp1cmd, x1, x2, y1, y2);
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_setres_raw(w: u32, h: u32, bpp: u32, rate: u32) {
	let (gp1cmd, x1, x2, y1, y2) = calc_gpu_res_params(w, h, bpp, rate);
	gp1(gp1cmd);
	gp1disprange(x1, x2, y1, y2);
}

#[cfg_attr(not(release), inline(never))]
pub fn ascii2sjis(chr: u8) -> u16 {
	match chr {
		0x00...0x1f => 0x88A0 + chr as u16 - 0x00,
		0x20 => 0x8140,
		0x21 => 0x8149,
		0x22 => 0x8168,
		0x23 => 0x8194,
		0x24 => 0x8190,
		0x25 => 0x8193,
		0x26 => 0x8195,
		0x27 => 0x8166,
		0x28 => 0x8169,
		0x29 => 0x816A, // prev+1
		0x2A => 0x8196,
		0x2B => 0x817B,
		0x2C => 0x8143,
		0x2D => 0x817C,
		0x2E => 0x8144,
		0x2F => 0x815E,
		0x30...0x39 => 0x824F + chr as u16 - 0x30,
		0x3A => 0x8146,
		0x3B => 0x8147, // prev+1
		0x3C => 0x8183,
		0x3D => 0x8181,
		0x3E => 0x8184,
		0x3F => 0x8148,
		0x40 => 0x8197,
		0x41...0x5A => 0x8260 + chr as u16 - 0x41,
		0x5B => 0x816D,
		0x5C => 0x818F,
		0x5D => 0x816E,
		0x5E => 0x814F,
		0x5F => 0x8151,
		0x60 => 0x8165,
		0x61...0x7A => 0x8281 + chr as u16 - 0x61,
		0x7B => 0x816F,
		0x7C => 0x8162,
		0x7D => 0x8170,
		0x7E => 0x8160,
		0x7F => 0x8189, // dummy
		0x80...0xff => 0x81F4, // dummy, should +chr-0x80 but no area
		_ => unreachable!()
	}
}

#[cfg_attr(not(release), inline(never))]
fn get_glyph(chr: u16) -> Option<(bool, &'static[u8])> {
	match chr {
		// FIXME using double-width-space glyph for single-width-space to make draw code simple, but semantic is completely broken.
		0x0020 => Option::Some((false, &sys::sjis2ptr(0x8140).unwrap()[0..15])),
		0x0021...0x007E => Option::Some((false, unsafe { std::slice::from_raw_parts((0xBFC7F8DE + (chr as u32 - 0x21) * 15) as *const u8, 15) })), // XXX touching inside BIOS but no other way!
		_ => sys::sjis2ptr(chr).and_then(|p| Option::Some((true, p))), // some ascii falls here that finally yields None.
	}
}

// TODO write 4bpp/8bpp ver? colorize?
#[cfg_attr(not(release), inline(never))]
pub fn draw_text_cpu_16bpp(buf: &mut [u16], width: usize, offx: usize, offy: usize, fg: u16, srx: usize, sry: usize, s: &[u16]) {
	let height = buf.len() / width;
	let mut x = offx;

	for c in s {
		let sjc = *c; // if *c < 0x100 { ascii2sjis(*c as u8) } else { *c };
		let (dw, gs) = get_glyph(sjc).expect("an unknown character in the text");
		let cc = if dw { 2 } else { 1 };
		let cw = if dw { 16 } else { 8 };
		// gs.len must be even if double width.
		/*
		for (y, bs) in gs.chunks(cc).take(height - offy).enumerate() {
			let line = if dw {
				((bs[0] as u32) << 8) | (bs[1] as u32)
			} else {
				(bs[0] as u32) << 8
			};
			for i in 0..cw {
				if (line << i) & 0x8000 != 0 {
					buf[width * (offy + y) + x + i] = fg;
				}
			}
		}
		*/
		let sch = gs.len() / cc; // Source Char Height
		let lpl = 1 << sry; // (source) Lines Per (scaled) Line
		for y in 0..cmp::min(roundup_div!(sch, lpl), height) {
			let mut line = [false; 16]; // max 16px width
			let py = y << sry;
			for yy in 0..cmp::min(lpl, sch - py) {
				let bi = (py + yy) * cc;
				let bs = &gs[bi..(bi+cc)];
				let lw = if dw {
					((bs[0] as u32) << 8) | (bs[1] as u32)
				} else {
					(bs[0] as u32) << 8
				};
				for i in 0..cw { // TODO is early-break faster? (in 0..1<<srx)
					line[i >> srx] |= ((lw << i) & 0x8000) != 0;
				}
			}
			for i in 0..(cw >> srx) {
				if line[i] {
					buf[width * (offy + y) + x + i] = fg;
				}
			}
		}
		x += cw >> srx;
	}
}

// TODO write 4bpp/8bpp ver? colorize?
#[cfg_attr(not(release), inline(never))]
pub fn get_text_pixmap_16bpp(fg: u16, bg: u16, srx: usize, sry: usize, s: &[u16]) -> (u16, u16, Box<[u16]>) {
	let bytes = s.iter().fold(0, |s, c| s + if *c < 0x100 { 1 } else { 2 } );
	let width = (8 >> srx) * bytes;
	let height = 16 >> sry;
	let mut buf = vec![bg; width * height].into_boxed_slice();
	draw_text_cpu_16bpp(&mut buf, width, 0, 0, fg, srx, sry, s);
	return (width as u16, height as u16, buf);
}

pub enum SemiTrans {
	Mix = 0, // B/2 + F/2, or implicitly express "Not Used"
	Add = 1, // B + F
	Sub = 2, // B - F
	AddQuarter = 3 // B + F/4
}

// FIXME too many args... use struct? but it embeds common parameters at compile time?
#[cfg_attr(not(release), inline(never))]
pub fn gpu_monotri(
		cmd: u8,
		rgb: (u8, u8, u8),
		v1: (i16, i16), // vN.xy: signed 11bit integer, -1024...+1023
		v2: (i16, i16),
		v3: (i16, i16)){
	gp0((cmd as u32) << 24 | ((rgb.2 as u32) << 16) | ((rgb.1 as u32) << 8) | ((rgb.0 as u32) << 0));
	gp0(((v1.1 as u16 as u32) << 16) | v1.0 as u16 as u32);
	gp0(((v2.1 as u16 as u32) << 16) | v2.0 as u16 as u32);
	gp0(((v3.1 as u16 as u32) << 16) | v3.0 as u16 as u32);
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_calc_texpage(
		tp: (u32, u32), // tp.x: 64px(1024/4bit), tp.y: 256px(512/1bit)
		strans: SemiTrans,
		tpbpp: u32 // 2bit, 4/8/15/resv bit
	) -> u32 {
	let dither: u32 = 1; // 1bit, 0=off(strip LSB) 1=dither. only for GP0.E1
	let drawmask: u32 = 1; // 1bit, 0=prohibit 1=allow. only for GP0.E1
	let texdis: u32 = 0; // 1bit, 0=tex-enable 1=tex-disable(with GP1.09)
	return (texdis << 11) | (drawmask << 10) | (dither << 9) | (tpbpp << 7) | ((strans as u32) << 5) | (tp.1 << 4) | tp.0;
}

// FIXME too many args... use struct? but it embeds common parameters at compile time?
#[cfg_attr(not(release), inline(never))]
pub fn gpu_textri(
		cmd: u8,
		rgb: (u8, u8, u8),
		v1: (i16, i16), // vN.xy: signed 11bit integer, -1024...+1023
		v2: (i16, i16),
		v3: (i16, i16),
		tp: (u32, u32), // tp.x: 64px(1024/4bit), tp.y: 256px(512/1bit)
		t1: (u8, u8), // tN.uv: (implicit) unsigned 8bit integer.
		t2: (u8, u8),
		t3: (u8, u8),
		strans: SemiTrans,
		tpbpp: u32, // 2bit, 4/8/15/resv bit
		clut: (u32, u32)) { // clut:: cx: 16px(1024/6bit) cy: 1px(512/9bit)
	let cluthw: u32 = ((clut.1 as u32) << 6) | clut.0 as u32;

	gp0((cmd as u32) << 24 | ((rgb.2 as u32) << 16) | ((rgb.1 as u32) << 8) | ((rgb.0 as u32) << 0));
	gp0(((v1.1 as u16 as u32) << 16) | v1.0 as u16 as u32);
	gp0((cluthw << 16) | ((t1.1 as u32) << 8) | (t1.0 as u32));
	gp0(((v2.1 as u16 as u32) << 16) | v2.0 as u16 as u32);
	gp0((gpu_calc_texpage(tp, strans, tpbpp) << 16) | ((t2.1 as u32) << 8) | (t2.0 as u32));
	gp0(((v3.1 as u16 as u32) << 16) | v3.0 as u16 as u32);
	gp0(((t3.1 as u32) <<  8) | t3.0 as u32); // upper 16bit is 0
}

/*
#[cfg_attr(not(release), inline(never))]
fn gpu_mono_quad(
		cmd: u8,
		rgb: (u8, u8, u8),
		v1: (i16, i16), // vN.xy: signed 11bit integer, -1024...+1023
		v2: (i16, i16),
		v3: (i16, i16),
		v4: (i16, i16)) {
	gpu_monotri(cmd, rgb, v1, v2, v3);
	gp0(((v4.1 as u16 as u32) << 16) | v4.0 as u16 as u32);
}
*/

// FIXME too many args... use struct? but it embeds common parameters at compile time?
#[cfg_attr(not(release), inline(never))]
pub fn gpu_texquad(
		cmd: u8,
		rgb: (u8, u8, u8),
		v1: (i16, i16), // vN.xy: signed 11bit integer, -1024...+1023
		v2: (i16, i16),
		v3: (i16, i16),
		v4: (i16, i16),
		tp: (u32, u32), // tp.x: 64px(1024/4bit), tp.y: 256px(512/1bit)
		t1: (u8, u8), // tN.uv: (implicit) unsigned 8bit integer.
		t2: (u8, u8),
		t3: (u8, u8),
		t4: (u8, u8),
		strans: SemiTrans,
		tpbpp: u32, // 2bit, 4/8/15/resv bit
		clut: (u32, u32)) { // clut:: cx: 16px(1024/6bit) cy: 1px(512/9bit)
	gpu_textri(cmd, rgb, v1, v2, v3, tp, t1, t2, t3, strans, tpbpp, clut);
	gp0(((v4.1 as u16 as u32) << 16) | v4.0 as u16 as u32);
	gp0(((t4.1 as u32) <<  8) | t4.0 as u32); // upper 16bit is 0
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_opq_raw_quad(
		v1: (i16, i16),
		v2: (i16, i16),
		v3: (i16, i16),
		v4: (i16, i16),
		tp: (u32, u32),
		t1: (u8, u8),
		t2: (u8, u8),
		t3: (u8, u8),
		t4: (u8, u8),
		tpbpp: u32, // 2bit, 4/8/15/resv bit
		clut: (u32, u32)) { // clut:: cx: 16px(1024/6bit) cy: 1px(512/9bit)
	gpu_texquad(0x2D, (0, 0, 0), v1, v2, v3, v4, tp, t1, t2, t3, t4, SemiTrans::Mix, tpbpp, clut)
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_opq_raw15_quad(
		v1: (i16, i16),
		v2: (i16, i16),
		v3: (i16, i16),
		v4: (i16, i16),
		tp: (u32, u32),
		t1: (u8, u8),
		t2: (u8, u8),
		t3: (u8, u8),
		t4: (u8, u8)) {
	gpu_opq_raw_quad(v1, v2, v3, v4, tp, t1, t2, t3, t4, 2, (0, 0))
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_opq_texblend_quad(
		rgb: (u8, u8, u8),
		v1: (i16, i16),
		v2: (i16, i16),
		v3: (i16, i16),
		v4: (i16, i16),
		tp: (u32, u32),
		t1: (u8, u8),
		t2: (u8, u8),
		t3: (u8, u8),
		t4: (u8, u8),
		tpbpp: u32, // 2bit, 4/8/15/resv bit
		clut: (u32, u32)) { // clut:: cx: 16px(1024/6bit) cy: 1px(512/9bit)
	gpu_texquad(0x2C, rgb, v1, v2, v3, v4, tp, t1, t2, t3, t4, SemiTrans::Mix, tpbpp, clut)
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_opq_texblend15_quad(
		rgb: (u8, u8, u8),
		v1: (i16, i16),
		v2: (i16, i16),
		v3: (i16, i16),
		v4: (i16, i16),
		tp: (u32, u32),
		t1: (u8, u8),
		t2: (u8, u8),
		t3: (u8, u8),
		t4: (u8, u8)){
	gpu_opq_texblend_quad(rgb, v1, v2, v3, v4, tp, t1, t2, t3, t4, 2, (0, 0))
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_semitrans_raw_quad(
		v1: (i16, i16),
		v2: (i16, i16),
		v3: (i16, i16),
		v4: (i16, i16),
		tp: (u32, u32),
		t1: (u8, u8),
		t2: (u8, u8),
		t3: (u8, u8),
		t4: (u8, u8),
		strans: SemiTrans,
		tpbpp: u32, // 2bit, 4/8/15/resv bit
		clut: (u32, u32)) { // clut:: cx: 16px(1024/6bit) cy: 1px(512/9bit)
	gpu_texquad(0x2F, (0, 0, 0), v1, v2, v3, v4, tp, t1, t2, t3, t4, strans, tpbpp, clut)
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_semitrans_raw15_quad(
		v1: (i16, i16),
		v2: (i16, i16),
		v3: (i16, i16),
		v4: (i16, i16),
		tp: (u32, u32),
		t1: (u8, u8),
		t2: (u8, u8),
		t3: (u8, u8),
		t4: (u8, u8),
		strans: SemiTrans) {
	gpu_semitrans_raw_quad(v1, v2, v3, v4, tp, t1, t2, t3, t4, strans, 2, (0, 0))
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_semitrans_texblend_quad(
		rgb: (u8, u8, u8),
		v1: (i16, i16),
		v2: (i16, i16),
		v3: (i16, i16),
		v4: (i16, i16),
		tp: (u32, u32),
		t1: (u8, u8),
		t2: (u8, u8),
		t3: (u8, u8),
		t4: (u8, u8),
		strans: SemiTrans,
		tpbpp: u32, // 2bit, 4/8/15/resv bit
		clut: (u32, u32)) { // clut:: cx: 16px(1024/6bit) cy: 1px(512/9bit)
	gpu_texquad(0x2E, rgb, v1, v2, v3, v4, tp, t1, t2, t3, t4, strans, tpbpp, clut)
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_semitrans_texblend15_quad(
		rgb: (u8, u8, u8),
		v1: (i16, i16),
		v2: (i16, i16),
		v3: (i16, i16),
		v4: (i16, i16),
		tp: (u32, u32),
		t1: (u8, u8),
		t2: (u8, u8),
		t3: (u8, u8),
		t4: (u8, u8),
		strans: SemiTrans) {
	gpu_semitrans_texblend_quad(rgb, v1, v2, v3, v4, tp, t1, t2, t3, t4, strans, 2, (0, 0))
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_line(cmd: u8, rgb: (u8, u8, u8), first: (i16, i16), last: (i16, i16)) {
	gp0(((cmd as u32) << 24) | ((rgb.2 as u32) << 16) | ((rgb.1 as u32) << 8) | ((rgb.0 as u32) << 0));
	gp0(((first.1 as u16 as u32) << 16) | ((first.0 as u16 as u32) << 0));
	gp0(((last.1 as u16 as u32) << 16) | ((last.0 as u16 as u32) << 0));
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_opq_mono_line(rgb: (u8, u8, u8), first: (i16, i16), last: (i16, i16)) {
	gpu_line(0x40, rgb, first, last);
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_polyline_start(cmd: u8, rgb: (u8, u8, u8)) {
	gp0(((cmd as u32) << 24) | ((rgb.2 as u32) << 16) | ((rgb.1 as u32) << 8) | ((rgb.0 as u32) << 0));
}

// TODO is GPU requires at least 2 verts?
#[cfg_attr(not(release), inline(never))]
pub fn gpu_mono_polyline<'a, T: IntoIterator<Item=&'a(i16, i16)>>(cmd: u8, rgb: (u8, u8, u8), points: T) {
	gpu_polyline_start(cmd, rgb);
	for v in points {
		gp0(((v.1 as u16 as u32) << 16) | ((v.0 as u16 as u32) << 0));
	}
	gp0(0x5555_5555); // magic termination code
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_opq_mono_polyline<'a, T: IntoIterator<Item=&'a(i16, i16)>>(rgb: (u8, u8, u8), points: T) {
	gpu_mono_polyline(0x48, rgb, points);
}

#[cfg_attr(not(release), inline(never))]
fn gpu_shaded_polyline<'a, T: IntoIterator<Item=&'a((i16, i16), (u8, u8, u8))>>(cmd: u8, points: T) {
	let mut iter = points.into_iter();
	match iter.next() {
		None => return,
		Some(&((x, y), rgb)) => {
			gpu_polyline_start(cmd, rgb);
			gp0(((y as u16 as u32) << 16) | ((x as u16 as u32) << 0));
			// continues...
		}
	}
	for &((x, y), (r, g, b)) in iter {
		gp0(((b as u32) << 16) | ((g as u32) << 8) | ((r as u32) << 0));
		gp0(((y as u16 as u32) << 16) | ((x as u16 as u32) << 0));
	}
	gp0(0x5555_5555); // magic termination code
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_opq_shaded_polyline<'a, T: IntoIterator<Item=&'a((i16, i16), (u8, u8, u8))>>(points: T) {
	gpu_shaded_polyline(0x58, points);
}

#[cfg_attr(not(release), inline(never))]
fn gpu_mono_rect(cmd: u8, (r, g, b): (u8, u8, u8), (left, top): (i16, i16), (width, height): (u16, u16)) {
	gp0(((cmd as u32) << 24) | ((b as u32) << 16) | ((g as u32) << 8) | r as u32);
	gp0(((top as u32) << 16) | left as u32);
	gp0(((height as u32) << 16) | width as u32);
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_opq_mono_rect(rgb: (u8, u8, u8), topleft: (i16, i16), widhei: (u16, u16)) {
	gpu_mono_rect(0x60, rgb, topleft, widhei);
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_semitrans_mono_rect(rgb: (u8, u8, u8), topleft: (i16, i16), widhei: (u16, u16)) {
	gpu_mono_rect(0x62, rgb, topleft, widhei);
}

// note: TexPage(GP0.E1) must be setup before calling this.
#[cfg_attr(not(release), inline(never))]
fn gpu_texrect(cmd: u8, (r, g, b): (u8, u8, u8),
		(left, top): (i16, i16),
		(width, height): (u16, u16),
		(tu, tv): (u8, u8), // t.uv: (implicit) unsigned 8bit integer.
		(cx, cy): (u32, u32)) { // clut:: cx: 16px(1024/6bit) cy: 1px(512/9bit)
	gp0(((cmd as u32) << 24) | ((b as u32) << 16) | ((g as u32) << 8) | r as u32);
	gp0(((top as u32) << 16) | left as u32);
	gp0(((cy as u32) << 22) | ((cx as u32) << 16) | ((tv as u32) << 8) | tu as u32);
	gp0(((height as u32) << 16) | width as u32);
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_opq_raw_rect(topleft: (i16, i16), widhei: (u16, u16), tuv: (u8, u8), clut: (u32, u32)) {
	gpu_texrect(0x65, (0, 0, 0), topleft, widhei, tuv, clut);
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_opq_raw15_rect(topleft: (i16, i16), widhei: (u16, u16), tuv: (u8, u8)) {
	gpu_texrect(0x65, (0, 0, 0), topleft, widhei, tuv, (0, 0));
}

#[cfg_attr(not(release), inline(never))]
pub fn gpu_semitrans_texblend_rect(rgb: (u8, u8, u8), topleft: (i16, i16), widhei: (u16, u16), tuv: (u8, u8), clut: (u32, u32)) {
	gpu_texrect(0x66, rgb, topleft, widhei, tuv, clut);
}

#[cfg_attr(not(release), inline(never))]
pub fn draw_text_slow(x: i16, y: i16, bgr: u16, scalex: u32, scaley: u32, s: &[u16]) {
	let vx = 640;
	let vy = 0;

	let (w, h, gbuf) = get_text_pixmap_16bpp(bgr, 0x0000, 0, 0, s);
	assert!(w < 256, "in draw_text_slow: GPU can only draw up to 256px width/height at once..."); // only for lazyness, should div
	gpu_load(vx, vy, w, h, &gbuf);
	let rw = (w as u32 * scalex / 0x10000) as i16;
	let rh = (h as u32 * scaley / 0x10000) as i16;
	let tpx = (vx >> 6) as u32;
	let tpy = (vy >> 8) as u32;
	let u = (vx & 0x3F) as u8; // 0x3F not 0xFF because tpx changes by 64px.
	let v = (vy & 0xFF) as u8;
	gpu_opq_raw15_quad(
		(x, y), (x+rw, y), (x, y+rh), (x+rw, y+rh),
		(tpx, tpy),
		(u, v), (u+w as u8, v), (u, v+h as u8), (u+w as u8, v+h as u8));
}

const CACHED_FONT_SRX: usize = 1;
const CACHED_FONT_SRY: usize = 1;
const CACHED_FONT_CW: usize = 16 >> CACHED_FONT_SRX;
const CACHED_FONT_CH: usize = 16 >> CACHED_FONT_SRY;
const CACHED_FONT_CHUNK: usize = 16;

#[cfg_attr(not(release), inline(never))]
pub fn cache_ascii_font(fg: u16) {
	for (i, cs) in (0x20..0x7F).collect::<Vec<_>>().chunks(CACHED_FONT_CHUNK).enumerate() {
		let sjcs = cs.iter().map(|c|ascii2sjis(*c)).collect::<Vec<u16>>();
		let (w, h, gbuf) = get_text_pixmap_16bpp(fg, 0x0000, CACHED_FONT_SRX, CACHED_FONT_SRY, &sjcs);
		gpu_load(896, (i * CACHED_FONT_CH) as u16, w, h, &gbuf);
	}
}

#[cfg_attr(not(release), inline(never))]
pub fn draw_ascii_text_cached(x: i16, y: i16, scalex: u32, scaley: u32, s: &[u8]) {
	// note: cached font must be aligned to texpage period. (for faster texuv calc.)
	let tpx = (896 >> 6) as u32;
	let tpy = (0 >> 8) as u32;
	let w = CACHED_FONT_CW;
	let h = CACHED_FONT_CH;
	let rh = (h as u32 * scaley / 0x10000) as i16;
	for (i, rc) in s.iter().enumerate() {
		let c = *rc as u32;
		let xx = ((w *  i     ) as u32 * scalex / 0x10000) as i16 + x;
		let rx = ((w * (i + 1)) as u32 * scalex / 0x10000) as i16 + x;
		let vx = (c - 0x20) % CACHED_FONT_CHUNK as u32 * w as u32;
		let vy = (c - 0x20) / CACHED_FONT_CHUNK as u32 * h as u32;
		let u = (vx & 0xFF) as u8;
		let v = (vy & 0xFF) as u8;
		gpu_opq_raw15_quad(
			(xx, y), (rx, y), (xx, y+rh), (rx, y+rh),
			(tpx, tpy),
			(u, v), (u+w as u8, v), (u, v+h as u8), (u+w as u8, v+h as u8));
	}
}

trait Scene {
	fn draw(&self);
	fn input(&mut self, bits: u16) -> Option<Box<Scene>>;
}

#[cfg_attr(not(release), inline(never))]
fn panic_hook(info: &std::panic::PanicInfo) {
	// I do not want to be disturbed by irq!
	sys::enter_critsect();

	let loc = info.location().unwrap();
	let file = loc.file();
	let line = loc.line();
	let msg = match info.payload().downcast_ref::<&'static str>() {
		Some(s) => *s,
		None => match info.payload().downcast_ref::<String>() {
			Some(s) => &s[..],
			None => "Box<Any>",
		}
	};

	gpu_setres_raw(3, 0, 0, 0);
	gp1(0x05000000 | (  0 << 0) | (0 << 10));
	gp0(0xE3000000 | (  0 << 0) | (0 << 10)); // draw(clip)x1y1
	gp0(0xE4000000 | (640 << 0) | ((0 + 240) << 10)); // draw(clip)x2y2
	gp0(0xE5000000 | (  0 << 0) | (0 << 11)); // drawoffset
	fillvram(255, 0, 0, 0, 8, 640, 8*4);

	cache_ascii_font(0xFFFF);
	draw_ascii_text_cached(0, 8*1, 0x10000, 0x10000, b"Panic");
	// FIXME I don't like malloc in panic_hook, but panic!(fmt, arg...) and format! do, so may be it is ok...?
	draw_ascii_text_cached(0, 8*3, 0x10000, 0x10000, &msg.bytes().collect::<Vec<u8>>());
	draw_ascii_text_cached(0, 8*4, 0x10000, 0x10000, &format!("at {}:{}", file, line).bytes().collect::<Vec<u8>>());

	loop{}
}

#[cfg_attr(not(release), inline(never))]
pub fn binread(fname: &[u8]) -> Result<Vec<u8>, sys::Errno> {
	let mut fd = try!(sys::file_open(fname, sys::ReadWrite::Read, sys::file_mode::SYNC));
	// FIXME allocate roundup(filesize, 2048), read at once, then shrink to actual size. maybe faster.
	//       but unknown to get filesize.
	//       to get filesize, don't use first_file, or you get stuck with NO$PSX 1.9 BIOS!
	let mut buf = Vec::new();
	loop {
		let off = buf.len();
		buf.append(&mut vec![0u8; 2048]); // Rust 1.10 allocates zeroes, memcpys that, then drop that.
		//buf.extend((0..2048).map(|_|0u8)); // Rust 1.10 extends byte by byte.
		match sys::file_read(&mut fd, &mut buf[off..]) {
			Ok(r) => {
				if r < 2048 {
					buf.truncate(off + r);
					return Result::Ok(buf);
				}
			},
			Err(err) => {
				// if align_error and read at least once, it indicates EOF (filesize aligned by sector size)
				if err == sys::Errno::AlignmentError && 0 < buf.len() {
					buf.truncate(off);
					return Result::Ok(buf);
				} else {
					return Result::Err(err);
				}
			}
		}
	}
}

pub struct GraphicLogger {
	y: i16,
}

impl GraphicLogger {
#[cfg_attr(not(release), inline(never))]
	pub fn new() -> GraphicLogger {
		fillvram(32, 32, 32, 0, 0, 640, 480);
		cache_ascii_font(0xFFFF);
		GraphicLogger {
			y: 16,
		}
	}

#[cfg_attr(not(release), inline(never))]
	pub fn writeln(&mut self, s: &[u8]) {
		draw_ascii_text_cached(0, self.y, 0x10000, 0x10000, s);
		self.y += CACHED_FONT_CH as i16;
	}
}

struct SPUMemAllocator {
	addr: u32,
}

impl SPUMemAllocator {
	#[cfg_attr(not(release), inline(never))]
	pub fn new() -> SPUMemAllocator {
		SPUMemAllocator {
			addr: 0x1000,
		}
	}

	#[cfg_attr(not(release), inline(never))]
	pub fn load_file(&mut self, path: &[u8]) -> u32 {
		// use unwrap_or_else instead of unwrap, to delaying from_utf8.
		// FIXME can't use unwrap_or_else??
		//       error[E0281]: type mismatch: the type `[closure@shootingscene.rs:22:41: 22:115 path:_]` implements the trait `std::ops::FnOnce<()>`, but the trait `std::ops::FnOnce<(sys::Errno,)>` is required (expected tuple, found ())
		//       let vb = binread(path).unwrap_or_else(||{ panic!("Can't load {}", ::std::str::from_utf8(path).unwrap()); vec![]});
		let vb = match binread(path) {
			Ok(vec) => vec,
			Err(_) => panic!("Can't load {}", ::std::str::from_utf8(path).unwrap())
		};
		let loadaddr = self.addr;
		spu_dmaload(loadaddr, &vb);
		self.addr += roundup!(vb.len() as u32, 16); // it should already aligned, because it is VB format.
		loadaddr
	}
}

#[cfg_attr(not(release), inline(never))]
#[no_mangle]
pub extern "C" fn main() {
	// Bootscreen left ExitFromException context as is, must call this or die on interrupt!
	// note: Bootscreen set StatusReg with interrupt-disabled, start_pad enables.
	sys::set_default_rfe();

	std::panic::set_hook(Box::new(panic_hook));

	let mut logger = GraphicLogger::new();
	logger.writeln(b"default_rfe OK");
	logger.writeln(b"set_hook OK");

	for i in 0..24 { spu_keyoff(i); } // FIXME want really stop, not release
	logger.writeln(b"spu_keyoff OK");

	sys::init_pad();
	logger.writeln(b"init_pad OK");
	sys::start_pad();
	logger.writeln(b"start_pad OK");

	let ev_vsync = match sys::open_event(sys::EventClass::VBlank, sys::EventSpec::Interrupted, Option::None) {
		Some(ev) => ev,
		None => panic!("could not open event for vblank")
	};
	sys::enable_event(ev_vsync);
	logger.writeln(b"open/enable_event OK");

	// mute SPU until scene loads sounds to SPU memory...
	spucnt_unmute(false);
	spu_vol((0x3FFF, 0x3FFF), (0x5EBC, 0x5EBC)); // but set default volume.

	// extra safe for XEBRA... but no effects.
	unsafe {
		volatile_store(0x1F801DB0 as *mut u16, 0x0000); // CD Left
		volatile_store(0x1F801DB2 as *mut u16, 0x0000); // CD Right
		volatile_store(0x1F801DB4 as *mut u16, 0x0000); // External Left
		volatile_store(0x1F801DB6 as *mut u16, 0x0000); // External Right
	}

	//let mut scene = testscene::TestScene::new(&mut logger);
	let mut scene: Box<Scene> = Box::new(flappyscene::FlappyScene::new());
	//let mut scene = clockscene::ClockScene::new(&mut logger);
	//let mut scene = pslogoscene::PSLogoScene::new(&mut logger);
	//let mut scene = td1scene::TD1Scene::new(&mut logger);
	//let mut scene: Box<Scene> = Box::new(shootingscene::ShootingScene::new());
	//let mut scene: Box<Scene> = Box::new(titlescene::TitleScene::new());
	/*
	logger.writeln(b"scene::new OK");
	logger.writeln(b"");
	logger.writeln(b"Press Circle to continue");
	while !sys::DigitalPadButtons::new(read_digipad()).circle() {}
	*/
	loop{
		scene.draw();

		let bits = read_digipad();
		prng::seed(prng::rand() + bits as u32);
		match scene.input(bits) {
			Some(newscene) => scene = newscene,
			None => (/*nothing*/)
		}

		// NO$PSX extension
		if unsafe{*(0x1F802060 as *const u8)} == 0x45 {
			while !sys::test_event(ev_vsync) {
				// NO$PSX: wait for next vsync with idleing.
				unsafe {
					volatile_store(0x1F802064 as *mut u8, 0x4F);
					volatile_store(0x1F802065 as *mut u8, 0x4E);
					volatile_load(0x1F802066 as *const u8);
				}
			}
			// test_event clears flag, no need to wait_event.
		} else {
			sys::wait_event(ev_vsync);
		}
	}
}
