//#![feature(asm)]

pub use self::imp::*;

use std::ops::Sub;

// TODO should remove "Q12" and support fastmath::Q
#[derive(Default, Clone, Copy)]
pub struct Vec3gQ12<T> {
	pub x: T, pub y: T, pub z: T
}
impl<T: Sub> Sub for Vec3gQ12<T> {
	type Output = Vec3gQ12<T::Output>;
	#[cfg_attr(not(release), inline(never))]
	fn sub(self, other: Vec3gQ12<T>) -> Vec3gQ12<T::Output> {
		return Vec3gQ12{x: self.x - other.x, y: self.y - other.y, z: self.z - other.z};
	}
}

pub type Vec3Q12 = Vec3gQ12<i16>;
pub type Vec3wQ12 = Vec3gQ12<i32>;

#[cfg(target_arch="mips")]
#[macro_use]
pub mod imp {
	use super::{Vec3Q12, Vec3wQ12};

	#[derive(Default, Clone, Copy)]
	pub struct Mat33gte {
		m1112: u32,
		m1321: u32,
		m2223: u32,
		m3132: u32,
		m33: i16,
	}

	// TODO impl Mat33tr that have q12 tr.
	#[derive(Default, Clone, Copy)]
	pub struct Mat33trgte {
		mat33: Mat33gte,
		tx: i32,
		ty: i32,
		tz: i32,
	}

	// should function, but it refuse be compiled with crate-type=lib because variable(reg) is invalid for "i"...
	/*
	fn mfc2(reg: u32) -> i32 {
		let dst: i32;
		unsafe {
			asm!("mfc2 $0, $$$1" : "=r"(dst) : "i"(reg));
		}
		return dst;
	}
	fn mtc2(reg: u32, val: i32) {
		unsafe {
			asm!("mtc2 $0, $$$1" : : "r"(val) "i"(reg));
		}
	}
	*/
	// FIXME rust 1.12.0 sometimes say 'used $at (currently $1) without ".set noat"' but $at is chosen by rust (or LLVM) itself!
	//       to quick-hacky avoid that, sand insn with ".set noat"..".set at".
	//       maybe it cause some nasty crash??
	// FIXME additional "nop" is not required if NOT followed by using $rd... but it requires LLVM Mips special.
	// FIXME this may not require be "volatile" (excl. LZCR), but without it sometimes rustc stop emitting [mc]fc2...
	macro_rules! mfc2 {
		($reg:expr) => {{
			let dst: i32;
			unsafe {
				if $reg < 32 {
					asm!(".set noat;mfc2 $0, $$$1;nop;.set at" : "=r"(dst) : "i"($reg) : : "volatile");
				} else {
					asm!(".set noat;cfc2 $0, $$$1;nop;.set at" : "=r"(dst) : "i"($reg - 32) : : "volatile");
				}
			}
			dst
		}}
	}
	// FIXME this may not require be "volatile" (excl. LZCS), but without it sometimes rustc stop emitting [mc]tc2...
	macro_rules! mtc2 {
		($reg:expr, $val:expr) => {{
			unsafe {
				if $reg < 32 {
					asm!(".set noat;mtc2 $0, $$$1;.set at" : : "r"($val), "i"($reg) : : "volatile");
				} else {
					asm!(".set noat;ctc2 $0, $$$1;.set at" : : "r"($val), "i"($reg - 32) : : "volatile");
				}
			}
		}}
	}
	// note: LLVM MIPS backend currently does not support COP2, use ".long".
	// FIXME preceeding "nop" is not required if this COP2 is not placed after mtc2/ctc2... but it requires LLVM Mips special.
	macro_rules! gte { // intended for internal use
		($op:expr, $sf:expr, $mtx:expr, $inv:expr, $trv:expr, $lm:expr) => {
			unsafe { asm!("nop;.long $0" : : "i"(0x4A000000 | $op | ($sf << 19) | ($mtx << 17) | ($inv << 15) | ($trv << 13) | ($lm << 10)) : : "volatile"); }
		}
	}
	macro_rules! gte_sf { // intended for internal use
		($op:expr, $sf:expr) => {
			gte!($op, $sf, 0, 0, 0, 0);
		}
	}
	macro_rules! mvmva { // in=Mx,Vx,Tx out=IRv/MACv
		($sf:expr, $mtx:expr, $inv:expr, $trv:expr, $sat:expr) => {
			gte!(0x0400012, $sf, $mtx, $inv, $trv, $sat);
		}
	}
	macro_rules! rtps { // in=V0,OFX,OFY,RTxx,TRx,DQA,DQB,H out=SXY2,SZ2,IR0 destroy=IRv,MAC0,MACv
		($sf:expr) => { gte_sf!(0x018_0001, $sf); }
	}
	macro_rules! rtpt { // in=V0,V1,V2,OFX,OFY,RTxx,TRx,DQA,DQB,H out=SXYn,SZn,IR0 destroy=IRv,MAC0,MACv
		($sf:expr) => { gte_sf!(0x028_0030, $sf); }
	}
	macro_rules! nclip { // in=SXY012 out=MAC0
		() => { gte_sf!(0x140_0006, 0); }
	}
	macro_rules! avsz3 { // in=ZSF3,SZ123 out=MAC0,OTZ
		() => { gte_sf!(0x150_002D, 1); }
	}
	macro_rules! sqr { // "IRv <- IRv . IRv" in=IRv out=IRv/MACv lm=1!
		($sf:expr) => { gte_sf!(0x0A0_0428, $sf); }
	}
	macro_rules! op { // "IRv <- Dv x IRv" in=IRv,Dv out=IRv/MACv
		($sf:expr) => { gte_sf!(0x170_000C, $sf); }
	}
	macro_rules! ncs { // in=LL,V0,BK,LC out=RGB2(,MACv,IRv) lm=1! XEBRA does not support lm=0!
		($sf:expr) => { gte_sf!(0x0C8_041E, $sf); }
	}
	macro_rules! nct { // in=LL,V012,BK,LC out=RGB012(,MACv,IRv) lm=1! XEBRA does not support lm=0!
		($sf:expr) => { gte_sf!(0x0D8_0420, $sf); }
	}
	macro_rules! nccs { // in=LL,V0,BK,LC,RGBC out=RGB2(,MACv,IRv) lm=1! XEBRA does not support lm=0!
		($sf:expr) => { gte_sf!(0x108_041B, $sf); }
	}
	macro_rules! ncct { // in=LL,V012,BK,LC,RGBC out=RGB012(,MACv,IRv) lm=1! XEBRA does not support lm=0!
		($sf:expr) => { gte_sf!(0x118_043F, $sf); }
	}
	macro_rules! ncds { // in=LL,V0,BK,LC,RGBC,FC,IR0 out=RGB2(,MACv,IRv) lm=1! XEBRA does not support lm=0!
		($sf:expr) => { gte_sf!(0x0E8_0413, $sf); }
	}
	macro_rules! ncdt { // in=LL,V012,BK,LC,RGBC,FC,IR0 out=RGB012(,MACv,IRv) lm=1! XEBRA does not support lm=0!
		($sf:expr) => { gte_sf!(0x0F8_0416, $sf); }
	}

	#[cfg_attr(not(release), inline(never))]
	pub fn get_mat33trgte_rt() -> Mat33trgte {
		Mat33trgte {
			mat33: Mat33gte {
				m1112: mfc2!(32) as u32,
				m1321: mfc2!(33) as u32,
				m2223: mfc2!(34) as u32,
				m3132: mfc2!(35) as u32,
				m33: mfc2!(36) as i16
			},
			tx: mfc2!(37),
			ty: mfc2!(38),
			tz: mfc2!(39)
		}
	}
	#[cfg_attr(not(release), inline(never))]
	pub fn load_mat33trgte_rt(mat: &Mat33trgte) {
		mtc2!(32, mat.mat33.m1112);
		mtc2!(33, mat.mat33.m1321);
		mtc2!(34, mat.mat33.m2223);
		mtc2!(35, mat.mat33.m3132);
		mtc2!(36, mat.mat33.m33 as i32);
		mtc2!(37, mat.tx);
		mtc2!(38, mat.ty);
		mtc2!(39, mat.tz);
	}
	/*pub fn mul_mat33trgte_gte(mat1: Mat33trgte, mat2: Mat33trgte) -> Mat33trgte {
		;
	}*/

	// TODO dedup mat9 and mat33
	macro_rules! load_mat9 { // intended for internal use
		($t:expr, $m:expr) => {
			{
				let m: &[i16; 9] = $m; // type check
				mtc2!($t * 8 + 32, ((m[1] as u16 as u32) << 16) | m[0] as u16 as u32);
				mtc2!($t * 8 + 33, ((m[3] as u16 as u32) << 16) | m[2] as u16 as u32);
				mtc2!($t * 8 + 34, ((m[5] as u16 as u32) << 16) | m[4] as u16 as u32);
				mtc2!($t * 8 + 35, ((m[7] as u16 as u32) << 16) | m[6] as u16 as u32);
				mtc2!($t * 8 + 36,                                m[8] as u16 as u32);
			}
		}
	}
	// non-macro (strict type check) yet assembler friendly in ugly way...
	#[cfg_attr(not(release), inline(never))] pub fn load_mat9_rt(m: &[i16; 9]) { load_mat9!(0, m) }
	#[cfg_attr(not(release), inline(never))] pub fn load_mat9_ll(m: &[i16; 9]) { load_mat9!(1, m) }
	#[cfg_attr(not(release), inline(never))] pub fn load_mat9_lc(m: &[i16; 9]) { load_mat9!(2, m) }

	#[cfg_attr(not(release), inline(never))]
	pub fn load_mat9_vs(m: &[i16; 9]) {
		mtc2!(0, ((m[3] as u16 as u32) << 16) | m[0] as u16 as u32); mtc2!(1, m[6] as u16 as u32);
		mtc2!(2, ((m[4] as u16 as u32) << 16) | m[1] as u16 as u32); mtc2!(3, m[7] as u16 as u32);
		mtc2!(4, ((m[5] as u16 as u32) << 16) | m[2] as u16 as u32); mtc2!(5, m[8] as u16 as u32);
	}
	macro_rules! mult_load_mat9 { // intended for internal use
		($t:expr, $m:expr) => {
			{
				let m: &[i16; 9] = $m; // type check
				// load matrix m as vectors
				$crate::mat::imp::load_mat9_vs(m);
				// multiply and collect (transposed) (old)*m
				mvmva!(1/*q12*/, $t, 0/*V0*/, 3/*None xlate.*/, 0/*s16 saturate*/);
				let t0 = (mfc2!(9) as i16, mfc2!(10) as i16, mfc2!(11) as i16);
				mvmva!(1/*q12*/, $t, 1/*V1*/, 3/*None xlate.*/, 0/*s16 saturate*/);
				let t1 = (mfc2!(9) as i16, mfc2!(10) as i16, mfc2!(11) as i16);
				mvmva!(1/*q12*/, $t, 2/*V2*/, 3/*None xlate.*/, 0/*s16 saturate*/);
				let t2 = (mfc2!(9) as i16, mfc2!(10) as i16, mfc2!(11) as i16);
				// load (old)*m with transpose
				load_mat9!($t, &[
								 t0.0, t1.0, t2.0,
								 t0.1, t1.1, t2.1,
								 t0.2, t1.2, t2.2,
				]);
			}
		}
	}
	// non-macro (strict type check) yet assembler friendly in ugly way...
	#[cfg_attr(not(release), inline(never))] pub fn mult_load_mat9_rt(m: &[i16; 9]) { mult_load_mat9!(0, m) }
	#[cfg_attr(not(release), inline(never))] pub fn mult_load_mat9_ll(m: &[i16; 9]) { mult_load_mat9!(1, m) }
	#[cfg_attr(not(release), inline(never))] pub fn mult_load_mat9_lc(m: &[i16; 9]) { mult_load_mat9!(2, m) }

	/// Vectors...

	macro_rules! gte_set_vec3 { ($base:expr, $vec:expr) => { // $vec must be Vec3Q12
		mtc2!($base + 0, $vec.x);
		mtc2!($base + 1, $vec.y);
		mtc2!($base + 2, $vec.z);
	} }
	macro_rules! gte_get_vec3 { ($base:expr) => {
		Vec3Q12 {
			x: mfc2!($base + 0) as i16,
			y: mfc2!($base + 1) as i16,
			z: mfc2!($base + 2) as i16
		}
	} }
	macro_rules! gte_set_vec2 { ($base:expr, $vec:expr) => { // $vec must be Vec3Q12
		mtc2!($base + 0, (($vec.y as i32) << 16) | ($vec.x as i32));
		mtc2!($base + 1, $vec.z);
	} }
	macro_rules! gte_get_vec2 { ($base:expr) => {
		{
			let xy = mfc2!($base + 0);
			Vec3Q12 {
				x: xy as i16,
				y: (xy >> 16) as i16,
				z: mfc2!($base + 1) as i16
			}
		}
	} }
	macro_rules! defvec3 { ($name:ident, $base:expr) => {
		pub mod $name {
			use super::super::Vec3Q12;
			#[cfg_attr(not(release), inline(never))]
			pub fn set(vec: Vec3Q12) { gte_set_vec3!($base, vec); }
			#[cfg_attr(not(release), inline(never))]
			pub fn get() -> Vec3Q12 { gte_get_vec3!($base) }
		}
	} }
	macro_rules! defvec2 { ($name:ident, $base:expr) => {
		pub mod $name {
			use super::super::Vec3Q12;
			#[cfg_attr(not(release), inline(never))]
			pub fn set(vec: Vec3Q12) { gte_set_vec2!($base, vec); }
			#[cfg_attr(not(release), inline(never))]
			pub fn get() -> Vec3Q12 { gte_get_vec2!($base) }
		}
	} }
	macro_rules! defvecm { ($name:ident, $base:expr) => {
		pub mod $name {
			use super::super::Vec3Q12;
			// setup "misused" Dvec
			#[cfg_attr(not(release), inline(never))]
			pub fn set_for_cross(vec: Vec3Q12) {
				mtc2!($base + 0, vec.x as u16 as u32);
				mtc2!($base + 2, vec.y as u16 as u32);
				mtc2!($base + 4, vec.z as u16);
			}
			// setup matrix for emulating dot-multiply
			#[cfg_attr(not(release), inline(never))]
			pub fn set_for_dot(vec: Vec3Q12) {
				set_for_cross(vec);
				mtc2!($base + 1, 0u32);
				mtc2!($base + 3, 0u32);
			}
		}
	} }
	defvec3!(ir, 9);
	defvec2!(v0, 0);
	defvecm!(rtv, 32);
	defvecm!(llv, 40);
	defvecm!(lcv, 48);

	///

	// TODO I can't define pow2 in generic...
	#[cfg_attr(not(release), inline(never))]
	fn pow2(x: u32, e: i32) -> u32 {
		if 0 <= e {
			x << e
		} else {
			x >> -e
		}
	}
	// input is uintq0 (or any q with adjusting rexp)
	// output is (mantissa_q12, exp) -> real value is mantissa_q12>>exp (q12)
	// mantissa have 12bit precision, 8bit(*3/4) resolution.
	// retval may be (u16,usize) but i16 is good for GTE vecs...
	#[cfg_attr(not(release), inline(never))]
	pub fn rsqrt16q12(x: u32) -> Option<(i16, usize)> {
		const RSR_TAB: [i16; 192] = [ // q12
			// 0b0...01... (1.0..2.0)
			0x1000,0xFE0,0xFC1,0xFA3,0xF85,0xF68,0xF4C,0xF30,0xF15,0xEFB,0xEE1,0xEC7,0xEAE,0xE96,0xE7E,0xE66,
			0xE4F,0xE38,0xE22,0xE0C,0xDF7,0xDE2,0xDCD,0xDB9,0xDA5,0xD91,0xD7E,0xD6B,0xD58,0xD45,0xD33,0xD21,
			0xD10,0xCFF,0xCEE,0xCDD,0xCCC,0xCBC,0xCAC,0xC9C,0xC8D,0xC7D,0xC6E,0xC5F,0xC51,0xC42,0xC34,0xC26,
			0xC18,0xC0A,0xBFD,0xBEF,0xBE2,0xBD5,0xBC8,0xBBB,0xBAF,0xBA2,0xB96,0xB8A,0xB7E,0xB72,0xB67,0xB5B,
			// 0b0...10... (2.0..3.0)
			0xB50,0xB45,0xB39,0xB2E,0xB24,0xB19,0xB0E,0xB04,0xAF9,0xAEF,0xAE5,0xADB,0xAD1,0xAC7,0xABD,0xAB4,
			0xAAA,0xAA1,0xA97,0xA8E,0xA85,0xA7C,0xA73,0xA6A,0xA61,0xA59,0xA50,0xA47,0xA3F,0xA37,0xA2E,0xA26,
			0xA1E,0xA16,0xA0E,0xA06,0x9FE,0x9F6,0x9EF,0x9E7,0x9E0,0x9D8,0x9D1,0x9C9,0x9C2,0x9BB,0x9B4,0x9AD,
			0x9A5,0x99E,0x998,0x991,0x98A,0x983,0x97C,0x976,0x96F,0x969,0x962,0x95C,0x955,0x94F,0x949,0x943,
			// 0b0...11... (3.0..4.0)
			0x93C,0x936,0x930,0x92A,0x924,0x91E,0x918,0x912,0x90D,0x907,0x901,0x8FB,0x8F6,0x8F0,0x8EB,0x8E5,
			0x8E0,0x8DA,0x8D5,0x8CF,0x8CA,0x8C5,0x8BF,0x8BA,0x8B5,0x8B0,0x8AB,0x8A6,0x8A1,0x89C,0x897,0x892,
			0x88D,0x888,0x883,0x87E,0x87A,0x875,0x870,0x86B,0x867,0x862,0x85E,0x859,0x855,0x850,0x84C,0x847,
			0x843,0x83E,0x83A,0x836,0x831,0x82D,0x829,0x824,0x820,0x81C,0x818,0x814,0x810,0x80C,0x808,0x804
		];

		mtc2!(30, x);
		unsafe { asm!("nop; nop" : : : : "volatile"); } // 1 cycle for mtc2 delay, and lzc may takes 1 extra cycle?
		let lzc = mfc2!(31);

		if lzc == 0 || lzc == 32 { // TODO can lzc==0 happen?
			return Option::None; // you can't get reciprocal value of zero by this function!
		}

		let lzc_even = lzc & !1; // evenize lzc to negative, this rsqrt operates on even-bit-pos (1.0..4.0).

		// idx = int(x * pow(2, lzc_even - 24)) - 0x40
		let iexp = lzc_even - 24;
		// normalizing x to ((1.0..4.0)-1.0)q6
		/*
		let idx = if 0 <= iexp {
			x << iexp
		} else {
			x >> (24 - lzc_even) /*=-iexp*/
		} - 0x40;
		*/
		let idx = pow2(x, iexp) - 0x40;
		debug_assert!(idx < 192);
		let vq12 = RSR_TAB[idx as usize];

		let rexp = (31 - lzc_even as usize) / 2;
		return Option::Some((vq12, rexp));
	}
	macro_rules! gte_normalize_ir {
		($v:ident, $i:expr) => { {
			sqr!(1); // q12; ir = ir * ir
			// calculate mac.x+mac.y+mac.z in the cpu (use mac instead of ir for accuracy)
			let ln = mfc2!(25) as u32 + mfc2!(26) as u32 + mfc2!(27) as u32;
			/*
			// or... use gte on all, but complex and less accuracy...
			// save irv to v0
			v0::set(ir::get());
			sqr!(1); // q12; ir *= ir
			// use mvmva to calculate ir.x+ir.y+ir.z
			mtc2!(32, 0x00010001);
			mtc2!(33, 0x00010000);
			mtc2!(34, 0);
			mtc2!(35, 0);
			mtc2!(36, 0);
			mvmva!(0, 0, 3, 3, 0); // q0,rt*ir+0
			let ln = mfc2!(9) as u32;
			*/

			let rsl = rsqrt16q12(ln).expect("Can't normalize zero-length vector"); // if None, it's program error, panic!
			$v::set_for_dot(Vec3Q12{x: rsl.0, y: rsl.0, z: rsl.0});
			mvmva!(1, $i, 3, 3, 0); // q12, $i*ir+0
			// /*if gte was used for getting ln...*/mvmva!(1, 0, 0, 3); // q12, rt*v0+0
			// adjust exp (note: ln*rsl.0>>rsl.1 is enough (no need to extra >>12).)
			let exp = pow2(0x1000, rsl.1 as i32 - 12) as i16;
			$v::set_for_dot(Vec3Q12{x: exp, y: exp, z: exp}); // if you need speed, use set_for_cross (because already zeroed r33/35).
			mvmva!(1, $i, 3, 3, 0); // q12, $i*ir+0
			/*
			// or... use cpu, but it costs.
			let rsl = ...;
			ir::set(Vec3Q12{
			x: ((v.x as u32 * rsl.0) >> rsl.1) as i16,
			y: ((v.y as u32 * rsl.0) >> rsl.1) as i16,
			z: ((v.z as u32 * rsl.0) >> rsl.1) as i16
			});
			*/
		} }
	}
	// non-macro (strict type check) yet assembler friendly in ugly way...
	#[cfg_attr(not(release), inline(never))] pub fn gte_normalize_ir_rt() { gte_normalize_ir!(rtv, 0) } // ir -> ir, destroy RT
	#[cfg_attr(not(release), inline(never))] pub fn gte_normalize_ir_ll() { gte_normalize_ir!(llv, 1) } // ir -> ir, destroy LL
	#[cfg_attr(not(release), inline(never))] pub fn gte_normalize_ir_lc() { gte_normalize_ir!(lcv, 2) } // ir -> ir, destroy LC

	#[cfg_attr(not(release), inline(never))]
	pub fn cpu_dot(v1: Vec3Q12, v2: Vec3Q12) -> i32 { // q12*q12->q12
		return (v1.x as i32 * v2.x as i32 + v1.y as i32 * v2.y as i32 + v1.z as i32 * v2.z as i32) >> 12;
	}

	// note: using function for normalize instead of macro, to easy debugging (no-inline).
	macro_rules! lookat_mat9tr {
		($m:ident, $normalizer:ident, $eye:expr, $at:expr, $up:expr) => { {
			let ema = $eye - $at;
			ir::set(ema);
			$normalizer();
			let z = ir::get();

			// keep IR hold z
			$m::set_for_cross($up);
			op!(1);
			$normalizer();
			let x = ir::get();

			$m::set_for_cross(z);
			op!(1);
			$normalizer();
			let y = ir::get();

			// TODO for dot, should use GTE (instead of cpu_dot)? "eye" can reused.
			//      (ie. set eye to RT and fetch xyz then mvmva and fetch xyz.eye... ugh, normalize destroys RT...)
			return ([ x.x, x.y, x.z, y.x, y.y, y.z, z.x, z.y, z.z ], [-cpu_dot(x, $eye), -cpu_dot(y, $eye), -cpu_dot(z, $eye)]);
		} }
	}
	// non-macro (strict type check) yet assembler friendly in ugly way...
	#[cfg_attr(not(release), inline(never))] pub fn lookat_mat9tr_rt(eye: Vec3Q12, at: Vec3Q12, up: Vec3Q12) -> ([i16; 9], [i32; 3]) { lookat_mat9tr!(rtv, gte_normalize_ir_rt, eye, at, up) }
	#[cfg_attr(not(release), inline(never))] pub fn lookat_mat9tr_ll(eye: Vec3Q12, at: Vec3Q12, up: Vec3Q12) -> ([i16; 9], [i32; 3]) { lookat_mat9tr!(llv, gte_normalize_ir_ll, eye, at, up) }
	#[cfg_attr(not(release), inline(never))] pub fn lookat_mat9tr_lc(eye: Vec3Q12, at: Vec3Q12, up: Vec3Q12) -> ([i16; 9], [i32; 3]) { lookat_mat9tr!(lcv, gte_normalize_ir_lc, eye, at, up) }

	#[cfg_attr(not(release), inline(never))]
	pub fn cpu_normalize32(v: Vec3wQ12) -> Vec3wQ12 { // q0->q12
		let ln = (v.x*v.x + v.y*v.y + v.z*v.z) as u32;
		let rsl = rsqrt16q12(ln).expect("Can't normalize zero-length vector");
		Vec3wQ12{
			x: (v.x * rsl.0 as i32) >> rsl.1,
			y: (v.y * rsl.0 as i32) >> rsl.1,
			z: (v.z * rsl.0 as i32) >> rsl.1
		}
	}

	#[cfg_attr(not(release), inline(never))]
	pub fn cpu_cross32(i: Vec3wQ12, d: Vec3wQ12) -> Vec3wQ12 { // q12 x qany -> qany
		Vec3wQ12 {
			x: (i.z * d.y - i.y * d.z) >> 12,
			y: (i.x * d.z - i.z * d.x) >> 12,
			z: (i.y * d.x - i.x * d.y) >> 12,
		}
	}

	#[cfg_attr(not(release), inline(never))]
	pub fn cpu_dot32(v1: Vec3wQ12, v2: Vec3wQ12) -> i32 { // q12*qany->qany TODO genericize?
		return (v1.x * v2.x + v1.y * v2.y + v1.z * v2.z) >> 12;
	}

	// eye, at and up should q0
	#[cfg_attr(not(release), inline(never))]
	pub fn cpu_lookat_mat9tr(eye: Vec3wQ12, at: Vec3wQ12, up: Vec3wQ12) -> ([i16;9], [i32;3]) {
		let z = cpu_normalize32(eye - at); // z become q12!
		let x = cpu_normalize32(cpu_cross32(z, cpu_normalize32(up))); // x become q12!
		let y = cpu_normalize32(cpu_cross32(x, z)); // y become q12!
		return ([
						x.x as i16, x.y as i16, x.z as i16,
						y.x as i16, y.y as i16, y.z as i16,
						z.x as i16, z.y as i16, z.z as i16 ],
						[-cpu_dot32(x, eye), -cpu_dot32(y, eye), -cpu_dot32(z, eye)]);
	}
}

// static mut matstack: [Mat33trgte; 10] = [Mat33trgte{..Default::default()}; 10];
/*
pub fn main() {
	/*

	unsafe {
		mat[2] = imp::get_mat();
		for i in 0..10 {
			imp::load_mat(mat[i]);
		}
	}
	*/
	// TODO perspective
	imp::lookat_mat33trgte(Vec3Q12{x:0x1000, y:0x2000, z:0x1000}, Vec3Q12{x: 0, y: 0, z: 0}, Vec3Q12{x: 0, y: -1, z: 0});
	// TODO rotate, translate, scale
	rtpt!();
}
*/
