use ::std::default::Default;
use ::std::cmp::{min, max};
use ::std::intrinsics::volatile_store;
use ::{Scene, SPUMemAllocator, binread, spucnt_unmute, spucnt_reverb, spu_vol, spu_play_ex};
use ::{SemiTrans, gpu_byteload, gpu_opq_raw15_rect, gp0, gp1, gpu_opq_mono_line, gpu_setres_raw, fillvram, set_draw_area, gpu_calc_texpage, gpu_semitrans_mono_rect}; // FIXME should moved to graphic
use ::progress::Progress;

pub struct TitleScene {
	btns: ::sys::DigitalPadButtonState,
	choir: u32,
	frame: u32,
	fade: i32,
}

impl TitleScene {
	pub fn new() -> TitleScene {
		let mut progress = Progress::new(2);
		let mut spu = SPUMemAllocator::new();

		let choir_addr = spu.load_file(b"cdrom:\\CHOIR.VB;1\0");
		progress.step();

		gpu_byteload(576, 0, 256, 64, binread(b"cdrom:\\MARK_15.RAW;1\0").expect("Can't load MARK_15.RAW"));
		progress.step();

		gpu_setres_raw(1, 0, 0, 0);

		// Space Echo
		unsafe {
			let params: [u16; 32] = [
				0x033D,0x0231,0x7E00,0x5000,0xB400,0xB000,0x4C00,0xB000,
				0x6000,0x5400,0x1ED6,0x1A31,0x1D14,0x183B,0x1BC2,0x16B2,
				0x1A32,0x15EF,0x15EE,0x1055,0x1334,0x0F2D,0x11F6,0x0C5D,
				0x1056,0x0AE1,0x0AE0,0x07A2,0x0464,0x0232,0x8000,0x8000
			];
			volatile_store(0x1F801DA2 as *mut u16, 0xE128);
			for (i,p) in params.iter().enumerate() {
				volatile_store((0x1F801DC0 + i * 2) as *mut u16, *p);
			}
		}
		spucnt_unmute(true);
		spucnt_reverb(true);
		spu_vol((0x3FFF, 0x3FFF), (0x3000, 0x3000));

		TitleScene {
			btns: Default::default(),
			choir: choir_addr,
			frame: 0,
			fade: 0,
		}
	}
}

impl Scene for TitleScene {
	fn draw(&self) {
		gp1(0x05000000 | (  0 << 0) | (0 << 10));

		gp0(0xE1000000 | gpu_calc_texpage((9, 0), SemiTrans::Sub, 2)); // texpage 9,0, Sub for fade
		gp0(0xE2000000); // texwindow
		gp0(0xE6000000); // maskbits
		set_draw_area((0, 0), (320-1, 0 + 240-1));
		fillvram(0, 0, 0, 0, 0, 320, 240);

		gpu_opq_raw15_rect((48, 50), (192, 48), (64, 0));
		gpu_opq_mono_line((251, 176, 14), (0, 100), (320, 100));
		gpu_opq_mono_line((251, 176, 14), (0, 101), (320, 101));
		if self.frame < 30 {
			gpu_opq_raw15_rect((48, 108), (192, 16), (64, 48));
		}

		if 0 < self.fade {
			let lumi = min(255, max(0, (self.fade - 15) * 255 / 45)) as u8;
			gpu_semitrans_mono_rect((lumi, lumi, lumi), (0, 0), (320, 240));
		}
	}
	fn input(&mut self, bits: u16) -> Option<Box<Scene>> {
		self.btns.update(bits);

		if 0 < self.fade {
			self.fade += 1;
			if self.fade == 70 {
				return Some(Box::new(::shootingscene::ShootingScene::new()));
			}
		}

		if self.btns.down().start() && self.fade == 0 {
			self.fade = 1;

			let env = 0x40008000; // TODO make more good
			spu_play_ex(0, self.choir, /*choir@D#3*/8154, (0x4000, 0x4000), env, true);
			spu_play_ex(1, self.choir, /*choir@G-3*/10255, (0x4000, 0x4000), env, true);
			spu_play_ex(2, self.choir, /*choir@G-4*/20511, (0x4000, 0x4000), env, true);
			spu_play_ex(3, self.choir, /*choir@A#4*/24360, (0x4000, 0x4000), env, true);
		}

		self.frame = (self.frame + 1) % 60;
		return None;
	}
}
