# coding: utf-8

class SNA
	Tag = Struct.new :tag, :zro, :data
	def initialize io
		read io
	end

	def read io
		@magic = io.read(16)
		@magic == "NO$PSX SNAPSHOT\x1a" or raise "?magic"
		@hdr = io.read(48)
		@chunks = {}
		until io.eof
			tag = io.read(4)
			zro = io.read(4)
			size = io.read(4).unpack("V")[0]
			data = io.read(size)
			@chunks[tag] = Tag.new tag, zro, data
			break if tag == "STOP"
		end
	end

	def write io
		io.write @magic
		io.write @hdr
		@chunks.each { |k, v|
			io.write v.tag
			io.write v.zro
			io.write [v.data.size].pack("V")
			io.write v.data
		}
	end

	def setregs xregs
		@chunks["CORE"].data[0x4004, 4*(31+3)] = xregs[0, 4*(31+3)] # r1-r31,lo,hi,pc
	end
	def setbios xbios #512K
		@chunks["BIOS"].data = xbios
	end
	def setmem xmem #2M
		@chunks["MRAM"].data = xmem
	end
	def setvram xvram #1M
		@chunks["VRAM"].data = xvram
	end
	def setsram xsram #512K
		@chunks["SRAM"].data = xsram
	end
end

# 86C-8EB r0-r31??
# 8EC-8EF lo
# 8F0-8F3 hi
# 8F4-8F7 pc
# 944-94F sr,cause,epc(cop0 r12-14)
# A14- GTE(cop2 r0-)?
# A94- DMA0-6_MADR,BCR,CHCR(planar),DPCR,DICR
# 3B5C-3BBF memctl?(1F801000-?)
# 12B5C-212B5B RAM?
# 212B5C-292B5B BIOS
# 2D2FA4- video??
# 2D3020- GP0cmd_FIFOs?
# 2D3108-3D3107 VRAM?
# 3D3108- sound?? 1F801C00-? \
# 3D3288- 1F801D80-?         } contiguous
# 3D32A8- 1F801DA0-?         /
# 3D3D44-453D43 SRAM??
# 45C058- MC*2?
class XebraImage
	def initialize io
		@data = io.read
	end

	def regs; @data[0x86C..0x8F7]; end
	def sr; @data[0x944, 4]; end
	def cause; @data[0x948, 4]; end
	def epc; @data[0x94C, 4]; end
	def ram; @data[0x12B5C..0x212B5B]; end
	def bios; @data[0x212B5C..0x292B5B]; end
	def vram; @data[0x2D2FA4, 0x100000]; end
	def sram; @data[0x3D3D44, 0x80000]; end
end

def main
	if ARGV[1]
		sna = open(ARGV[0], "rb") { |f| SNA.new f }
		xim = open(ARGV[1], "rb") { |f| XebraImage.new f }
		sna.setregs xim.regs
		sna.setbios xim.bios
		sna.setmem xim.ram
		sna.setvram xim.vram
		open(ARGV[2], "wb") { |f| sna.write f }
	else
		xim = open(ARGV[0], "rb") { |f| XebraImage.new f }
		puts "r00-r07: #{xim.regs[4* 0, 4*8].unpack('V*').map{|e| '%08X' % e }.join(' ')}"
		puts "r08-r15: #{xim.regs[4* 8, 4*8].unpack('V*').map{|e| '%08X' % e }.join(' ')}"
		puts "r16-r23: #{xim.regs[4*16, 4*8].unpack('V*').map{|e| '%08X' % e }.join(' ')}"
		puts "r24-r31: #{xim.regs[4*24, 4*8].unpack('V*').map{|e| '%08X' % e }.join(' ')}"
		puts "lo, hi, pc: #{xim.regs[4*32, 4*3].unpack('V*').map{|e| '%08X' % e }.join(' ')}"
		puts "epc: %08X" % xim.epc.unpack('V*')[0]
	end
end

$0 == __FILE__ and main
