# QEMU PlayStation State Loader
#
# build:
#   $ mipsel-linux-musl-as --mips32 -o xl.o xloader.s && mipsel-linux-musl-ld -o xl -Ttext 0xbfc00000 -e start xl.o && mipsel-linux-musl-objdump -dr xl | less
#   $ mipsel-linux-musl-objcopy -O binary -j .text xl xl.bin && hexdump -C xl.bin
# use:
#   > ruby xebra_st.rb foo.state mipsel_bios.bin
#   > qemu-system-mipsel -net none -vga none -k nul -monitor stdio -gdb tcp::23946 -S
#   $ mipsel-linux-musl-gdb -q game.elf -ex 'target remote QEMUMACHINE:23946' -ex 'tb *0xbfc000fc' -ex c -ex 'set $pc=*(int*)0xbfc00188' -ex 'lay split' -ex 'set directories $cwd:/path/to/rust'

	.text

	.set reorder
	.set noat

start:
	# make COP0 usable 1/2
	mfc0	$at, $12 # c0_status
	li	$t0, 0x10000000 # Cu0
	or	$at, $t0
	b	start2
	. = 0x10
	.long 0xEEEEeeee # a word at 0x10 will be destructed by qemu

start2:
	# make COP0 usable 2/2
	mtc0	$at, $12 # c0_status

	# setup TLB to make same as PSX
	#mtc0	$zero, $0 #c0_index # inited to 0 by qemu
	#mtc0	$zero, $10 #c0_entryhi # inited to 0 by qemu
	li	$at, 0x003FF000 # 4MB page
	mtc0	$at, $5 #c0_pagemask
	li	$at, 0x00000017 # PFN=0 C=2(uncached) D=1 V=1 G=1
	mtc0	$at, $2 #c0_entrylo0
	mtc0	$at, $3 #c0_entrylo1
	tlbwi

	# transfer memory
	li	$t0, 0x80000000 # dst
	li	$t2, 0x80200000 # dst's end
	li	$t1, 0xbfc80000 # src
ramloop:
	lw	$t3, 0($t1)
	sw	$t3, 0($t0)
	addiu	$t0, $t0, 4
	addiu	$t1, $t1, 4
	bne	$t0, $t2, ramloop

	# transfer LO/HI
	lw	$at, 0x0180($ra)
	mflo	$at
	lw	$at, 0x0184($ra)
	mfhi	$at
	# 0x018C: PC

	# transfer COP0 regs
	lw	$at, 0x0190($ra)
	mtc0	$at, $12 #c0_sr
	lw	$at, 0x0194($ra)
	mtc0	$at, $13 #c0_cause
	lw	$at, 0x0198($ra)
	mtc0	$at, $14 #c0_epc

	# transfer GPRs
	lui	$ra, 0xbfc0
	lw	$at, 0x0104($ra)
	lw	$v0, 0x0108($ra)
	lw	$v1, 0x010c($ra)
	lw	$a0, 0x0110($ra)
	lw	$a1, 0x0114($ra)
	lw	$a2, 0x0118($ra)
	lw	$a3, 0x011c($ra)
	lw	$t0, 0x0120($ra)
	lw	$t1, 0x0124($ra)
	lw	$t2, 0x0128($ra)
	lw	$t3, 0x012c($ra)
	lw	$t4, 0x0130($ra)
	lw	$t5, 0x0134($ra)
	lw	$t6, 0x0138($ra)
	lw	$t7, 0x013c($ra)
	lw	$s0, 0x0140($ra)
	lw	$s1, 0x0144($ra)
	lw	$s2, 0x0148($ra)
	lw	$s3, 0x014c($ra)
	lw	$s4, 0x0150($ra)
	lw	$s5, 0x0154($ra)
	lw	$s6, 0x0158($ra)
	lw	$s7, 0x015c($ra)
	lw	$t8, 0x0160($ra)
	lw	$t9, 0x0164($ra)
	lw	$k0, 0x0168($ra)
	lw	$k1, 0x016c($ra)
	lw	$gp, 0x0170($ra)
	lw	$sp, 0x0174($ra)
	lw	$s8, 0x0178($ra)
	lw	$ra, 0x017c($ra)

	b	.

	#.rept 0x100 - (. - start)
	#nop
	#.endr
